/**
 ****************************************************************************************
 *
 * @file dataflash.c
 *
 * @brief DATAFLASH driver
 *
 * Copyright(C) 2012-2014 Renesas Electronics Corporation
 *
 * rev.205
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup DATAFLASH
 *
 * @{
 ****************************************************************************************
 */

/*
 * COMPILER SETTING
 ****************************************************************************************
 */

/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include "r_cg_macrodriver.h"
#include "dataflash.h"
#include "fdl.h"                    /* FAL(FDL) Library Header    */
#include "eel.h"                    /* EEL      Library Header    */
#include "fdl_descriptor.h"         /* FAL(FDL) Descriptor Header */
#include "eel_descriptor.h"         /* EEL      Descriptor Header */
#include "r_cg_userdefine.h"

/*
 * DEFINES
 ****************************************************************************************
 */
typedef struct df_errno
{
    uint8_t eel;
    uint8_t df ;
} df_errno_t;

typedef enum
{
    DF_STAT_UNINIT = 0,
    DF_STAT_STOP  ,
    DF_STAT_START ,
    DF_STAT_PROC  
} df_stat_t;

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */
static df_stat_t        df_stat;    /* driver status    */
static eel_command_t    df_cmd;     /* driver command   */
static eel_request_t    df_req;     /* driver parameter */

static const df_errno_t errno_list[] =
{
    {(uint8_t)EEL_OK                   , DF_OK                   },
    {(uint8_t)EEL_BUSY                 , DF_OK                   },
    {(uint8_t)EEL_ERR_INITIALIZATION   , DF_ERR_INITIALIZATION   },
    {(uint8_t)EEL_ERR_ACCESS_LOCKED    , DF_ERR_ACCESS_LOCKED    },
    {(uint8_t)EEL_ERR_VERIFY           , DF_ERR_COMMAND          },
    {(uint8_t)EEL_ERR_PARAMETER        , DF_ERR_PARAMETER        },
    {(uint8_t)EEL_ERR_REJECTED         , DF_ERR_REJECTED         },
    {(uint8_t)EEL_ERR_NO_INSTANCE      , DF_ERR_NO_INSTANCE      },
    {(uint8_t)EEL_ERR_POOL_FULL        , DF_ERR_POOL_FULL        },
    {(uint8_t)EEL_ERR_POOL_INCONSISTENT, DF_ERR_POOL_INCONSISTENT},
    {(uint8_t)EEL_ERR_POOL_EXHAUSTED   , DF_ERR_POOL_EXHAUSTED   },
    {(uint8_t)EEL_ERR_INTERNAL         , DF_ERR_INTERNAL         }
};

extern LoRa_Param_Flag_t    lpf;

/*
 * FUNCTION PROTOTYPES
 ****************************************************************************************
 */
static uint8_t dataflash_errno(eel_status_t eel_errno);
static void dataflash_busy_handler(void);

/*
 * FUNCTION DEFINITIONS
 ****************************************************************************************
 */
/**
 ****************************************************************************************
 * @brief   initialize dataflash library(EEL/FAL)
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_init(void)
{
    uint8_t         errno = DF_ERR_UNKNOWN;
    fdl_status_t    fal_status; /* parameter for FAL */

    /* initialize parameter for EEL */
    memset(&df_req, 0, sizeof(df_req));

    /* initialize FAL */
    fal_status = FDL_Init(&fdl_descriptor_str);
    if(fal_status == FDL_OK)
    {
        /* Enable access for Data Flash memory */
        FDL_Open();

        /* initialize EEL */
        df_req.status_enu = EEL_Init();
        if(df_req.status_enu == EEL_OK)
        {
            /* if initialization succeeded, change state */
            df_stat = DF_STAT_STOP;
        }
        errno = dataflash_errno(df_req.status_enu);
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   open and start dataflash library(EEL/FAL)
 *
 * @param[in] mode      execution mode
 *                       - DF_MODE_ENFORCED
 *                          continue to execute until processing finished.
 *                          when DataFlash is not formatted, execute to format.
 *                       - DF_MODE_POLLING
 *                          start to processing.
 *                          need to call "dataflash_continue" until processing finished.
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_start(const uint8_t mode)
{
    uint8_t errno = DF_ERR_DISALLOWED;

    if(df_stat == DF_STAT_STOP)
    {
        /* -- OPEN -- (state : closed -> opened) */
        EEL_Open();

        /* -- START -- (state : opened -> started) */
        df_cmd = EEL_CMD_STARTUP;
        df_req.command_enu = EEL_CMD_STARTUP;
        EEL_Execute(&df_req);

        if(mode == DF_MODE_ENFORCED)
        {
            dataflash_busy_handler();

            /* when DataFlash block is not initialized */
            if(df_req.status_enu == EEL_ERR_POOL_INCONSISTENT)
            {
                /* execute "EEL_CMD_FORMAT" command */
                df_req.command_enu = EEL_CMD_FORMAT;
                EEL_Execute(&df_req);

                dataflash_busy_handler();

                if(df_req.status_enu == EEL_OK)
                {
                    /* re-execute "EEL_CMD_STARTUP" command */
                    df_req.command_enu = EEL_CMD_STARTUP;
                    EEL_Execute(&df_req);
                    dataflash_busy_handler();
                }
            }
        }

        errno = dataflash_errno(df_req.status_enu);
        if(errno == DF_OK)
        {
            if(mode == DF_MODE_ENFORCED)
            {
                /* if start processing succeeded, change state */
                df_stat = DF_STAT_START;
            }
            else /* mode == DF_MODE_POLLING */
            {
                /* it is neccesary to execute "dataflash_continue", change state */
                df_stat = DF_STAT_PROC;
            }
        }
        else
        {
            /* do nothing */
        }
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   stop and close dataflash library(EEL/FAL)
 *
 * @param[in ] mode     execution mode
 *                       - DF_MODE_ENFORCED
 *                          continue to execute until processing finished.
 *                          when DataFlash is not formatted, execute to format.
 *                       - DF_MODE_POLLING
 *                          start to processing.
 *                          need to call "dataflash_continue" until processing finished.
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_stop(const uint8_t mode)
{
    uint8_t errno = DF_ERR_DISALLOWED;

    if(df_stat == DF_STAT_START)
    {
        /* -- STOP -- (state : started -> opened) */
        df_cmd = EEL_CMD_SHUTDOWN;
        df_req.command_enu = EEL_CMD_SHUTDOWN;
        EEL_Execute(&df_req);

        errno = dataflash_errno(df_req.status_enu);
        if(errno == DF_OK)
        {
            if(mode == DF_MODE_ENFORCED)
            {
                /* -- CLOSE -- (state : opened -> closed) */
                EEL_Close();

                /* if stop processing succeeded, change state */
                df_stat = DF_STAT_STOP;
            }
            else /* mode == DF_MODE_POLLING */
            {
                /* it is neccesary to execute "dataflash_continue", change state */
                df_stat = DF_STAT_PROC;
            }
        }
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   read data from DataFlash / write data to DataFlash
 *
 * @param[in ] mode     execution mode
 *                       - DF_MODE_ENFORCED
 *                          continue to execute until processing finished.
 *                          when DataFlash is not formatted, execute to format.
 *                       - DF_MODE_POLLING
 *                          start to processing.
 *                          need to call "dataflash_continue" until processing finished.
 * @param[in ] rw       read or write (DF_READ/DF_WRITE)
 * @param[in ] id       data id(registered to EEL in order to specify data)
 * @param[out] addr     destination RAM address
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_rw(const uint8_t mode, const df_rw_t rw, const uint8_t id, uint8_t* addr)
{
    uint8_t errno;

    if(df_stat == DF_STAT_START)
    {
        df_cmd = (eel_command_t)((rw == DF_READ) ? (EEL_CMD_READ) : (EEL_CMD_WRITE));
        df_req.address_pu08   = addr;
        df_req.identifier_u08 = id;
        df_req.command_enu    = df_cmd;
        EEL_Execute(&df_req);

        errno = dataflash_errno(df_req.status_enu);
        if(errno == DF_OK)
        {
            if(mode == DF_MODE_POLLING)
            {
                /* it is neccesary to execute "dataflash_continue", change state */
                df_stat = DF_STAT_PROC;
            }
            else
            {
                dataflash_busy_handler();
            }
        }
    }
    else if(df_stat == DF_STAT_PROC)
    {
        errno = DF_ERR_BUSY;
    }
    else
    {
        errno = DF_ERR_DISALLOWED;
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   cleanup DataFlash
 *
 * @param[in ] mode     execution mode
 *                       - DF_MODE_ENFORCED
 *                          continue to execute until processing finished.
 *                          when DataFlash is not formatted, execute to format.
 *                       - DF_MODE_POLLING
 *                          start to processing.
 *                          need to call "dataflash_continue" until processing finished.
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_cleanup(const uint8_t mode)
{
    uint8_t errno;

    if(df_stat == DF_STAT_START)
    {
        df_cmd = EEL_CMD_REFRESH;
        df_req.command_enu = EEL_CMD_REFRESH;
        EEL_Execute(&df_req);

        errno = dataflash_errno(df_req.status_enu);
        if(errno == DF_OK)
        {
            if(mode == DF_MODE_POLLING)
            {
                /* it is neccesary to execute "dataflash_continue", change state */
                df_stat = DF_STAT_PROC;
            }
            else
            {
                dataflash_busy_handler();
            }
        }
    }
    else if(df_stat == DF_STAT_PROC)
    {
        errno = DF_ERR_BUSY;
    }
    else
    {
        errno = DF_ERR_DISALLOWED;
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   format DataFlash
 *
 * @param[in ] mode     execution mode
 *                       - DF_MODE_ENFORCED
 *                          continue to execute until processing finished.
 *                          when DataFlash is not formatted, execute to format.
 *                       - DF_MODE_POLLING
 *                          start to processing.
 *                          need to call "dataflash_continue" until processing finished.
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_format(const uint8_t mode)
{
    uint8_t errno;

    if((df_stat == DF_STAT_STOP) || (df_stat == DF_STAT_START))
    {
        if(df_stat == DF_STAT_STOP)
        {
            EEL_Open();
        }
        df_cmd = EEL_CMD_FORMAT;
        df_req.command_enu = EEL_CMD_FORMAT;
        EEL_Execute(&df_req);

        errno = dataflash_errno(df_req.status_enu);
        if(errno == DF_OK)
        {
            if(mode == DF_MODE_ENFORCED)
            {
                dataflash_busy_handler();

                /* -- CLOSE -- (state : opened -> closed) */
                EEL_Close();

                /* if format finished, change state */
                df_stat = DF_STAT_STOP;
            }
            else /* mode == DF_MODE_POLLING */
            {
                /* it is neccesary to execute "dataflash_continue", change state */
                df_stat = DF_STAT_PROC;
            }
        }
    }
    else if(df_stat == DF_STAT_PROC)
    {
        errno = DF_ERR_BUSY;
    }
    else
    {
        errno = DF_ERR_DISALLOWED;
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   continue to execute processing
 *
 * @param[in] fin_flg   flag indicating whether processing finished or not
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_continue(bool* fin_flg)
{
    uint8_t errno = DF_ERR_DISALLOWED;

    if(df_stat == DF_STAT_PROC)
    {
        if( df_req.status_enu == EEL_BUSY )
        {
            EEL_Handler();
        }
        /* EEL_Execute retains the pointer of df_req, and EEL_Handler updates df_req */
        errno = dataflash_errno(df_req.status_enu);

        if( df_req.status_enu != EEL_BUSY )
        {
            /* processing finished */
            *fin_flg = true;

            switch( df_cmd )
            {
                case EEL_CMD_SHUTDOWN:
                case EEL_CMD_FORMAT:
                    /* -- CLOSE -- (state : opened -> closed) */
                    EEL_Close();

                    /* if below processing finished, pass here */
                    /*     -> STOP, FORMAT                     */
                    df_stat = DF_STAT_STOP;
                    break;
                default:
                    /* if below processing finished, pass here */
                    /*     -> START, READ, WRITE, CLEANUP      */
                    df_stat = DF_STAT_START;
                    break;
            }
        }
        else
        {
            /* processing unfinished */
            *fin_flg = false;
        }
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   get size of available space
 *
 * @param[out] wsize    size of available space
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_get_space(uint16_t* wsize)
{
    uint8_t errno = DF_ERR_DISALLOWED;
    eel_status_t status;

    if(df_stat == DF_STAT_START)
    {
        status = EEL_GetSpace((eel_u16*)wsize);
        *wsize = (status == EEL_OK) ? (*wsize) : (0);

        errno = dataflash_errno(status);
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   get EEL version string
 *
 * @param[out] ver      version string buffer
 *
 * @return  uint8_t     error code
 ****************************************************************************************
 */
uint8_t dataflash_getver(uint8_t* ver)
{
    uint8_t  errno = DF_ERR_DISALLOWED;
    uint8_t  len;
    uint8_t __far*  string;

    if((df_stat == DF_STAT_STOP) || (df_stat == DF_STAT_START))
    {
        string = EEL_GetVersionString();

        /* copy string to buffer */
        for(len = 0 ; len < 24; len++)
        {
            ver[len] = string[len];
            if(ver[len] == '\0')
            {
                break;
            }
        }
        /* fill by zero after end of string */
        for(len = (uint8_t)(len + 1); len < 24; len++)
        {
            ver[len] = '\0';
        }
        errno = DF_OK;
    }

    return errno;
}

/**
 ****************************************************************************************
 * @brief   inform whether dataflash is stopped or not
 *          (if dataflash is started, prohibit STOP/HALT of MCU)
 *
 * @return  bool    true: stopped, false: not stopped (started)
 ****************************************************************************************
 */
bool dataflash_is_stopped(void)
{
    return ((df_stat == DF_STAT_UNINIT) || (df_stat == DF_STAT_STOP)) ? (true) : (false);
}

/**
 ****************************************************************************************
 * @brief   convert from eel error code to dataflash driver error code
 *
 * @param[in] eel_errno     eel error code
 *
 * @return  uint8_t         error code
 ****************************************************************************************
 */
static uint8_t dataflash_errno(const eel_status_t eel_errno)
{
    uint8_t idx;

    for(idx = 0; idx < (sizeof(errno_list) / sizeof(errno_list[0])); idx++)
    {
        if(errno_list[idx].eel == eel_errno)
        {
            return errno_list[idx].df;
        }
    }
    return DF_ERR_UNKNOWN;
}

/**
 ****************************************************************************************
 * @brief   execute EEL_Handler for busy state (EEL Pack02)
 *
 * @return  none
 ****************************************************************************************
 */
static void dataflash_busy_handler(void)
{
    while(df_req.status_enu == EEL_BUSY)
    {
        EEL_Handler();
    }
}

uint8_t dataflash_w_lora_param_flag(LoRa_Param_Flag_t *buf)
{
    uint8_t ret = DF_ERR_UNKNOWN;

    /* try to get device address from DataFlash */
    if(dataflash_start(DF_MODE_ENFORCED) == DF_OK)
    {
        /* get BD_address(Public Device Address) from DataFlash */
        ret = dataflash_rw(DF_MODE_ENFORCED, DF_WRITE, EEL_ID_LORA_PARAM_FLAG, (uint8_t*)buf);

        dataflash_stop(DF_MODE_ENFORCED);
    }

    return ret;
}

uint8_t dataflash_r_lora_param_flag(LoRa_Param_Flag_t *buf)
{
    uint8_t ret = DF_ERR_UNKNOWN;

    /* try to get device address from DataFlash */
    if(dataflash_start(DF_MODE_ENFORCED) == DF_OK)
    {
        /* get BD_address(Public Device Address) from DataFlash */
        ret = dataflash_rw(DF_MODE_ENFORCED, DF_READ, EEL_ID_LORA_PARAM_FLAG, (uint8_t *)buf);

        dataflash_stop(DF_MODE_ENFORCED);
    }

    return ret;
}

void dataflash_read_param(void)
{
    uint8_t ret;

    ret = dataflash_r_lora_param_flag(&lpf);

    if ((ret != DF_OK) && (ret != DF_ERR_NO_INSTANCE))
    {
        /* data falsh error */
        LED1 = LED_ON;
        LED2 = LED_ON;
        __stop();
        while(1);
    }
}

void dataflash_erase_param(void)
{
    memset(&lpf, 0, sizeof(LoRa_Param_Flag_t));

    dataflash_w_lora_param_flag(&lpf);

    LED1 = LED_ON;
    LED2 = LED_ON;
    __stop();
    while(1);
}
