/******************************************************************************
 * DISCLAIMER
 * This software is supplied by Renesas Electronics Corporation and is only
 * intended for use with Renesas products. No other uses are authorized. This
 * software is owned by Renesas Electronics Corporation and is protected under
 * all applicable laws, including copyright laws.
 * THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
 * THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
 * LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
 * TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS
 * ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE
 * FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
 * ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * Renesas reserves the right, without notice, to make changes to this
 * software and to discontinue the availability of this software. By using this
 * software, you agree to the additional terms and conditions found by
 * accessing the following link:
 * http://www.renesas.com/disclaimer
 *
 * Copyright (C) 2019 Renesas Electronics Corporation. All rights reserved.
 *****************************************************************************/
/*******************************************************************************
* File Name    : r_ms8607.c
* Version      : 1.0
* Description  : Device driver for PHT Combination Sensor MS8607 of
*              :                                   TE Connectivity company.
*              : (PHT: pressure, humidity and temperature)
******************************************************************************/
/*****************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 31.05.2019 1.00     First Release
******************************************************************************/

/******************************************************************************
Includes
******************************************************************************/

#include "r_cg_macrodriver.h"
#include "r_cg_serial.h"
#include "r_cg_timer.h"
#include "types.h"
#include "r_ms8607.h"
#include "r_cg_userdefine.h"

/******************************************************************************
Typedef definitions
******************************************************************************/

/******************************************************************************
Macro definitions
******************************************************************************/

#define IICA_START_CONDITION_WAIT_COUNT     100

/******************************************************************************
Exported global variables (to be accessed by other files)
******************************************************************************/

/* The ADC converison time. Used in r_usleep() function.    */
/* Also used in r_cg_timer_user.c as a decrement counter. */
int16_t wait_timer_count;

/******************************************************************************
Private global variables and functions
******************************************************************************/

/* Pressure(D1) Read Command Table */
static const uint8_t d1_cmd[6] = {
    MS8607_I2C_CMD_CONVERT_D1_OSR_256,
    MS8607_I2C_CMD_CONVERT_D1_OSR_512,
    MS8607_I2C_CMD_CONVERT_D1_OSR_1024,
    MS8607_I2C_CMD_CONVERT_D1_OSR_2048,
    MS8607_I2C_CMD_CONVERT_D1_OSR_4096,
    MS8607_I2C_CMD_CONVERT_D1_OSR_8192
};

/* Temperature(D2) Read Command Table */
static const uint8_t d2_cmd[6] = {
    MS8607_I2C_CMD_CONVERT_D2_OSR_256,
    MS8607_I2C_CMD_CONVERT_D2_OSR_512,
    MS8607_I2C_CMD_CONVERT_D2_OSR_1024,
    MS8607_I2C_CMD_CONVERT_D2_OSR_2048,
    MS8607_I2C_CMD_CONVERT_D2_OSR_4096,
    MS8607_I2C_CMD_CONVERT_D2_OSR_8192
};

/* Pressure and Temperature Convertion Wait Time Table */
static const uint8_t d1d2_conv_delay[6] = {
    MS8607_OSR_256_CONV_DELAY_MS,
    MS8607_OSR_512_CONV_DELAY_MS,
    MS8607_OSR_1024_CONV_DELAY_MS,
    MS8607_OSR_2048_CONV_DELAY_MS,
    MS8607_OSR_4096_CONV_DELAY_MS,
    MS8607_OSR_8192_CONV_DELAY_MS
};

/* Humedity Convertion Wait Time Table */
static const uint8_t hum_conv_delay[4] = {
    MS8607_12B_CONV_DELAY_MS,
    MS8607_8B_CONV_DELAY_MS,
    MS8607_10B_CONV_DELAY_MS,
    MS8607_11B_CONV_DELAY_MS
};

/* Calibration coefficient values */
static uint32_t                     g_ms8607_prom_coeffs[7];
/* Humedity ADC resolution */
static ms8607_humidity_resolution_t g_ms8607_humidity_res;
/* Presure ADC resolution */
static ms8607_pressure_resolution_t g_ms8607_pressure_res;

static int8_t r_CRC16(uint8_t *);
static void r_usleep(int16_t msec);

void R_IIC_Transfer_End(void);
MD_STATUS R_IIC_Send(uint8_t adr, uint8_t * const tx_buf, uint16_t tx_num);
MD_STATUS R_IIC_Receive(uint8_t adr, uint8_t * const rx_buf, uint16_t rx_num);

/******************************************************************************
* Function Name: R_MS8607_Init
* Description  : Initialize the MS8607 ADC resolution variables.
*                And initialize the IICA0 varibles.
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_MS8607_Init(void)
{
    /* set ADC resolution */
    g_ms8607_pressure_res = MS8607_PRESSURE_RESOLUTION_OSR_8192;
    g_ms8607_humidity_res = MS8607_HUMIDITY_RESOLUTION_12B;

    /* clear tranfer complete flag */
    /* used in r_cg_serial.c and r_cg_serial_user.c*/
    g_iic_trans_comp_flag = 0;

    return;
}

/******************************************************************************
* Function Name: R_MS8607_Reset
* Description  : Send I2C Reset command to MS8607.
* Arguments    : void
* Return Value : MS8607_STATUS_OK - 
*                    Success
*                MS8607_STATUS_I2C_TRANSFER_ERROR - 
*                    I2C transfer error
******************************************************************************/
ms8607_status R_MS8607_Reset(void)
{
	uint8_t     tx_buf[1];
    MD_STATUS   ret;

    /* Send I2C reset command */
    tx_buf[0] = MS8607_I2C_CMD_RESET;
    ret = R_IIC_Send(MS8607_I2C_PT_ADDR, tx_buf, 1);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* Send I2C humidity reset command */
    tx_buf[0] = MS8607_I2C_CMD_HUM_RESET;
    ret = R_IIC_Send(MS8607_I2C_HUM_ADDR, tx_buf, 1);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

	return MS8607_STATUS_OK;
}

/******************************************************************************
* Function Name: R_MS8607_Read_Prom
* Description  : Read factory calibrated coefficients from PROM
* Arguments    : void
* Return Value : MS8607_STATUS_OK - 
*                    Success
*                MS8607_STATUS_I2C_TRANSFER_ERROR - 
*                    I2C transfer error
******************************************************************************/
ms8607_status R_MS8607_Read_Prom(void)
{
    uint8_t     tx_buf[1] = {0};
    uint8_t     rx_buf[2] = {0};
    int8_t      i;
    MD_STATUS   ret;

    for (i = 0; i < 7; i++)
    {
        tx_buf[0] = MS8607_I2C_CMD_READ_PROM_BASE + (2 * i);
        /* command send */
        ret = R_IIC_Send(MS8607_I2C_PT_ADDR, (uint8_t*)tx_buf, 1);
        if (ret != MD_OK)
        {
            return MS8607_STATUS_I2C_TRANSFER_ERROR;
        }

        /* read 2bytes*/
        ret = R_IIC_Receive(MS8607_I2C_PT_ADDR, rx_buf, 2);
        if (ret != MD_OK)
        {
            return MS8607_STATUS_I2C_TRANSFER_ERROR;
        }

        /* read 7bytes */
        g_ms8607_prom_coeffs[i] = (uint32_t)((uint32_t)rx_buf[0] << 8)
                                  + (uint32_t)rx_buf[1];
    }

    return MS8607_STATUS_OK;
}

/******************************************************************************
* Function Name: R_MS8607_Set_Humidity_Resolution
* Description  : Read the user register from the device, modify its
*                contents to reflect the resolution, and then write 
*                the updated user register value to the device.
* Arguments    : void
* Return Value : MS8607_STATUS_OK - 
*                    Success
*                MS8607_STATUS_I2C_TRANSFER_ERROR - 
*                    I2C transfer error
******************************************************************************/
ms8607_status R_MS8607_Set_Humidity_Resolution(void)
{
    uint8_t     tx_buf[2];
    uint8_t     rx_buf[1];
    MD_STATUS   ret;

    /* Read user register */
    tx_buf[0] = MS8607_I2C_CMD_READ_USER_REG;
    ret = R_IIC_Send(MS8607_I2C_HUM_ADDR, tx_buf, 1);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    ret = R_IIC_Receive(MS8607_I2C_HUM_ADDR, rx_buf, 1);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* Modify user register to reflect resolution change */
    tx_buf[1] = rx_buf[0] & ~(MS8607_RESOLUTION_BIT7_MASK | MS8607_RESOLUTION_BIT0_MASK); /* Zero out bits 7 and 0 */

    if((g_ms8607_humidity_res == MS8607_HUMIDITY_RESOLUTION_10B) ||
       (g_ms8607_humidity_res == MS8607_HUMIDITY_RESOLUTION_11B))
    {
        tx_buf[1] |= MS8607_RESOLUTION_BIT7_MASK; /* Set bit 7 */
    }
    if((g_ms8607_humidity_res == MS8607_HUMIDITY_RESOLUTION_8B) ||
       (g_ms8607_humidity_res == MS8607_HUMIDITY_RESOLUTION_11B))
    {
        tx_buf[1] |= MS8607_RESOLUTION_BIT0_MASK; /* Set bit 0 */
    }

    /* Write user register */
    tx_buf[0] = MS8607_I2C_CMD_WRITE_USER_REG;
    ret = R_IIC_Send(MS8607_I2C_HUM_ADDR, tx_buf, 2);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    return MS8607_STATUS_OK;
}

/******************************************************************************
* Function Name: R_MS8607_Read_Temperature_Pressure_Humidity
* Description  : Send I2C commands to start a temperature conversion,
*                wait for completion, read the temperature value, start
*                a pressure conversion, wait for completion, read the
*                pressure value, start a relative humidity conversion,
*                wait for completion, and read the relative humidity value
* Arguments    : float* temperature -
*                    pointer to temperature variable
*                float* pressure -
*                    pointer to pressure variable
*                float* humidity -
*                    pointer to relative humidity variable
* Return Value : MS8607_STATUS_OK - 
*                    Success
*                MS8607_STATUS_I2C_TRANSFER_ERROR - 
*                    I2C transfer error
*                MS8607_STATUS_CRC_ERROR - 
*                    CRC error of the relative humidity
******************************************************************************/
ms8607_status R_MS8607_Read_Temperature_Pressure_Humidity(float* temperature, float* pressure, float* relative_humidity)
{
    uint8_t         tx_buf[1];
    uint8_t         rx_buf[3];
    uint32_t        adc16;
    uint32_t        pres_raw;
    uint32_t        temp_raw;
    int32_t         dT;
    int32_t         TEMP;
    long long int   OFF;
    long long int   SENS;
    int32_t         P;
    float           humidity;
    MD_STATUS       ret;

    *temperature       = 0;
    *pressure          = 0;
    *relative_humidity = 0;
    
    /***************************************************/
    /************* Pressure ADC Conversion *************/
    /***************************************************/
    /* send start pressure ADC conversion command */
    tx_buf[0] = d1_cmd[g_ms8607_pressure_res];
    ret = R_IIC_Send(MS8607_I2C_PT_ADDR, tx_buf, 1);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* wait for ADC conversion */
    r_usleep(d1d2_conv_delay[g_ms8607_pressure_res]);

    /* send read pressure ADC command */
    tx_buf[0] = MS8607_I2C_CMD_READ_ADC;
    ret = R_IIC_Send(MS8607_I2C_PT_ADDR, tx_buf, 1);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* receive pressure ADC result */
    ret = R_IIC_Receive(MS8607_I2C_PT_ADDR, rx_buf, 3);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    pres_raw = (uint32_t)((uint32_t)rx_buf[0] << 16)
               + (uint32_t)((uint32_t)rx_buf[1] << 8)
               + (uint32_t)(rx_buf[2]);

    /***************************************************/
    /*********** Temperature ADC Conversion ************/
    /***************************************************/
    /* send start temperature ADC conversion command */
    tx_buf[0] = d2_cmd[g_ms8607_pressure_res];
    ret = R_IIC_Send(MS8607_I2C_PT_ADDR, tx_buf, 1);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* wait for ADC conversion */
    r_usleep(d1d2_conv_delay[g_ms8607_pressure_res]);

    /* send read temperature ADC command */
	tx_buf[0] = MS8607_I2C_CMD_READ_ADC;
    ret = R_IIC_Send(MS8607_I2C_PT_ADDR, tx_buf, 1);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* receive temperature ADC result */
    ret = R_IIC_Receive(MS8607_I2C_PT_ADDR, rx_buf, 3);
    if(ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    temp_raw = (uint32_t)((uint32_t)rx_buf[0] << 16)
               + (uint32_t)((uint32_t)rx_buf[1] << 8)
               + (uint32_t)(rx_buf[2]);

    /* Calculate pressure and temperature using calibration coefficient values. */
    dT = temp_raw - (g_ms8607_prom_coeffs[5] << 8);
    TEMP = (uint32_t)2000 + (((long long int)dT * ((long long int)g_ms8607_prom_coeffs[6])) >> 23);
    OFF = (((long long int)g_ms8607_prom_coeffs[2]) << 17) + ((((long long int)g_ms8607_prom_coeffs[4]) * dT) >> 6);
    SENS = (((long long int)g_ms8607_prom_coeffs[1]) << 16) + ((((long long int)g_ms8607_prom_coeffs[3]) * dT) >> 7);
    P = (int32_t)((((((long long int)pres_raw) * SENS) >> 21) - OFF) >> 15);

    *pressure = (float)P / 100.0f;
    *temperature = (float)TEMP / 100.0f;

    /***************************************************/
    /******** Relative Humidity ADC Conversion *********/
    /***************************************************/
    /* send start Humidity ADC Conversion Command */
    tx_buf[0] = MS8607_I2C_CMD_MEAS_HUM_WITHOUT_HOLD;
    ret = R_IIC_Send(MS8607_I2C_HUM_ADDR, tx_buf, 1);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* wait only as long as is needed for the resolution that is currently set */
    r_usleep(hum_conv_delay[g_ms8607_humidity_res]);

    /* read humidity ADC result */
    ret = R_IIC_Receive(MS8607_I2C_HUM_ADDR, rx_buf, 3);
    if (ret != MD_OK)
    {
        return MS8607_STATUS_I2C_TRANSFER_ERROR;
    }

    /* CRC relative humidity data */
    if(r_CRC16(rx_buf) == TRUE)
    {
        /* Concatenate the received bytes into the 16 bit result */
        adc16 = (rx_buf[0] << 8) + (rx_buf[1] & 0xF0);
        /* Use formula to convert ADC result to relative humidity as a percentage */
        humidity = (float)(-6.0) + (float)(125.0) * (float)adc16 / (float)(65536.0);
        /* Bound humidity from 0% to 100% */
        if(humidity < 0)
        {
            humidity = 0;
        }
        else if(humidity > 100)
        {
            humidity = 100;
        }
    }
    else
    {
        return MS8607_STATUS_CRC_ERROR;
    }

    /* Assign relative humidity */
    *relative_humidity = humidity;

    return MS8607_STATUS_OK;
}

/******************************************************************************
* Function Name: r_CRC16
* Description  : Check 2 bytes of data with 8 bits of CRC information.
*                Use polynomial X^8 + X^5 + X^4 + 1
*                For use with TSYS02D serial number read and checking
*                ADC results for several sensors
* Arguments    : uint8_t* data -
*                    3 byte character array. data[0] is the most significant
*                    byte, data[1] is the least significant byte, and data[2]
*                    contains the CRC information.
* Return Value : TRUE -
*                    success
*                FALSE -
*                    error
******************************************************************************/
static int8_t r_CRC16(uint8_t *data)
{
    uint32_t    div, poly;
    int8_t      i;
    int8_t      ret = FALSE;

    div = (uint32_t)((uint32_t)data[0] << 16)
          + (uint32_t)((uint32_t)data[1] << 8)
          + (uint32_t)data[2];

    poly = CRC_POLY;

    for (i = 0; i < 16; i++)
    {
        if (((uint32_t)1 << (23 - i)) & div)
        {
            div ^= (poly << (15 - i));
        }
    }
    if ((div & (0xFF)) == 0x00)
    {
        ret = TRUE;
    }
    
    return ret;
}

/******************************************************************************
* Function Name: r_usleep
* Description  : Wait for the ADC conversion complete.
*                Uses the TAU0 Channel0 of RL78
* Arguments    : int16_t* msec -
*                    wait time.
* Return Value : void
******************************************************************************/
static void r_usleep(int16_t msec)
{
    wait_timer_count = msec; /* msec */

    R_TAU0_Channel0_Start();
    /* Decrement of wait_timer_count is in r_tau0_channel0_interrupt() */
    while(wait_timer_count > 0);
    R_TAU0_Channel0_Stop();
}

/******************************************************************************
* Function Name: R_IIC_Send
* Description  : Wrapper function of IIC send function.
* Arguments    : uint8_t adr -
*                    Slave address.
*                uint8_t * const tx_buf -
*                    Pointer to send data buffer.
*                uint16_t tx_num -
*                    Number of send data.
* Return Value : MD_OK - 
*                    Success.
*                MD_ERROR1 - 
*                    At the start of transmission, the bus is busy.
*                MD_ERROR2 - 
*                    Start condition is not detected.
*                MD_ERROR3 - 
*                    Bus release was time out.
******************************************************************************/
MD_STATUS R_IIC_Send(uint8_t adr, uint8_t * const tx_buf, uint16_t tx_num)
{
    MD_STATUS   ret;

    g_iic_trans_comp_flag = IIC_TRANSFER_NOT_YET;
    ret = R_IICA0_Master_Send(adr, tx_buf, tx_num, IICA_START_CONDITION_WAIT_COUNT);
    if (ret == MD_OK)
    {
        while (g_iic_trans_comp_flag == IIC_TRANSFER_NOT_YET);
        if (g_iic_trans_comp_flag == IIC_TRANSFER_COMPLETE)
        {
            ret = MD_OK;
        }
        else
        {
            ret = MD_ERROR3;/* BUS RELEASE TIME OUT */
        }
    }
    return (ret);
}

/******************************************************************************
* Function Name: R_IIC_Receive
* Description  : Wrapper function of IIC receive function.
* Arguments    : uint8_t adr -
 *                   Slave address.
*                uint8_t * const rx_buf -
*                    Pointer to receive data buffer.
*                uint16_t rx_num -
*                    Number of receive data.
* Return Value : MD_OK - 
*                    Success.
*                MD_ERROR1 - 
*                    At the start of transmission, the bus is busy.
*                MD_ERROR2 - 
*                    Start condition is not detected.
*                MD_ERROR3 - 
*                    Bus release was time out.
******************************************************************************/
MD_STATUS R_IIC_Receive(uint8_t adr, uint8_t * const rx_buf, uint16_t rx_num)
{
    MD_STATUS   ret;

    g_iic_trans_comp_flag = IIC_TRANSFER_NOT_YET;
    ret = R_IICA0_Master_Receive(adr, rx_buf, rx_num, IICA_START_CONDITION_WAIT_COUNT);
    if (ret == MD_OK)
    {
        while (g_iic_trans_comp_flag == IIC_TRANSFER_NOT_YET);
        if (g_iic_trans_comp_flag == IIC_TRANSFER_COMPLETE)
        {
            ret = MD_OK;
        }
        else
        {
            ret = MD_ERROR3;/* BUS RELEASE TIME OUT */
        }
    }
    return (ret);
}

/******************************************************************************
* Function Name: R_IIC_Transfer_End
* Description  : Wait for end of transfer of IIC.
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_IIC_Transfer_End(void)
{
    while(g_iic_trans_comp_flag == IIC_TRANSFER_NOT_YET);
    g_iic_trans_comp_flag = IIC_TRANSFER_NOT_YET;

    return;
}
