/******************************************************************************
 * DISCLAIMER
 * This software is supplied by Renesas Electronics Corporation and is only
 * intended for use with Renesas products. No other uses are authorized. This
 * software is owned by Renesas Electronics Corporation and is protected under
 * all applicable laws, including copyright laws.
 * THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
 * THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
 * LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
 * TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS
 * ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE
 * FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
 * ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * Renesas reserves the right, without notice, to make changes to this
 * software and to discontinue the availability of this software. By using this
 * software, you agree to the additional terms and conditions found by
 * accessing the following link:
 * http://www.renesas.com/disclaimer
 *
 * Copyright (C) 2019 Renesas Electronics Corporation. All rights reserved.
 *****************************************************************************/
/*******************************************************************************
* File Name    : r_ms8607.h
* Version      : 1.0
* Description  : Device driver for PHT Combination Sensor MS8607 of
*              :                                   TE Connectivity company.
*              : (PHT: pressure, humidity and temperature)
******************************************************************************/
/*****************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 31.05.2019 1.00     First Release
******************************************************************************/

#ifndef MS8607_H
#define MS8607_H

/******************************************************************************
Exported global variables
******************************************************************************/

/******************************************************************************
Macro definitions
******************************************************************************/

/* MS8607 Constants */
#define MS8607_I2C_PT_ADDR              (0x76<<1)   // 1110_110x (Pressure and Temperature)
#define MS8607_I2C_HUM_ADDR             (0x40<<1)   // 1000_000x (Relative Humidity)
#define MS8607_OSR_8192_CONV_DELAY_MS   (18)        // Datasheet lists max conversion time of 17.2ms, 18ms to be safe
#define MS8607_OSR_4096_CONV_DELAY_MS   (9)         // Datasheet lists max conversion time of 8.61ms, 9ms to be safe
#define MS8607_OSR_2048_CONV_DELAY_MS   (5)         // Datasheet lists max conversion time of 4.32ms, 5ms to be safe
#define MS8607_OSR_1024_CONV_DELAY_MS   (3)         // Datasheet lists max conversion time of 2.17ms, 3ms to be safe
#define MS8607_OSR_512_CONV_DELAY_MS    (2)         // Datasheet lists max conversion time of 1.06ms, 2ms to be safe
#define MS8607_OSR_256_CONV_DELAY_MS    (1)         // Datasheet lists max conversion time of 0.56ms, 1ms to be safe
#define MS8607_12B_CONV_DELAY_MS        (16)        // Datasheet lists max conversion time of 15.89ms for 12 bit resolution
#define MS8607_11B_CONV_DELAY_MS        (9)         // Datasheet lists max conversion time of 8.03ms for 11 bit resolution
#define MS8607_10B_CONV_DELAY_MS        (5)         // Datasheet lists max conversion time of 4.08ms for 10 bit resolution
#define MS8607_8B_CONV_DELAY_MS         (3)         // Datasheet lists max conversion time of 2.12ms for 8 bit resolution
#define CRC_POLY                        (0x131)     // CRC Polynomial: X^8 + X^5 + X^4 + 1

/* MS8607 BIT MASKS */
#define MS8607_RESOLUTION_BIT7_MASK             (0x80)
#define MS8607_RESOLUTION_BIT0_MASK             (0x01)
#define MS8607_BATTERY_STATUS_MASK              (0x40)
#define MS8607_HEATER_STATUS_MASK               (0x04)

/* MS8607 I2C Commands */
#define MS8607_I2C_CMD_RESET                    (0x1E)
#define MS8607_I2C_CMD_HUM_RESET                (0xFE)
#define MS8607_I2C_CMD_READ_PROM_BASE           (0xA0)
/* Pressure */
#define MS8607_I2C_CMD_CONVERT_D1_OSR_256       (0x40)
#define MS8607_I2C_CMD_CONVERT_D1_OSR_512       (0x42)
#define MS8607_I2C_CMD_CONVERT_D1_OSR_1024      (0x44)
#define MS8607_I2C_CMD_CONVERT_D1_OSR_2048      (0x46)
#define MS8607_I2C_CMD_CONVERT_D1_OSR_4096      (0x48)
#define MS8607_I2C_CMD_CONVERT_D1_OSR_8192      (0x4A)
/* Temperature */
#define MS8607_I2C_CMD_CONVERT_D2_OSR_256       (0x50)
#define MS8607_I2C_CMD_CONVERT_D2_OSR_512       (0x52)
#define MS8607_I2C_CMD_CONVERT_D2_OSR_1024      (0x54)
#define MS8607_I2C_CMD_CONVERT_D2_OSR_2048      (0x56)
#define MS8607_I2C_CMD_CONVERT_D2_OSR_4096      (0x58)
#define MS8607_I2C_CMD_CONVERT_D2_OSR_8192      (0x5A)
#define MS8607_I2C_CMD_READ_ADC                 (0x00)
/* Humidity */
#define MS8607_I2C_CMD_MEAS_HUM_WITHOUT_HOLD    (0xF5)
#define MS8607_I2C_CMD_WRITE_USER_REG           (0xE6)
#define MS8607_I2C_CMD_READ_USER_REG            (0xE7)

/******************************************************************************
Typedef definitions
******************************************************************************/

/* Error status */
typedef enum ms8607_status {
    MS8607_STATUS_OK                    = 0,
    MS8607_STATUS_I2C_TRANSFER_ERROR    = 1,
    MS8607_STATUS_CRC_ERROR             = 2,
    MS8607_STATUS_HEATER_ON_ERROR       = 3
} ms8607_status;

/* Measurement resolution */
typedef enum ms8607_humidity_resolution {
    MS8607_HUMIDITY_RESOLUTION_12B      = 0,    // 4096
    MS8607_HUMIDITY_RESOLUTION_8B       = 1,    //  256
    MS8607_HUMIDITY_RESOLUTION_10B      = 2,    // 1024
    MS8607_HUMIDITY_RESOLUTION_11B      = 3     // 2048
} ms8607_humidity_resolution_t;

typedef enum ms8607_pressure_resolution {
    MS8607_PRESSURE_RESOLUTION_OSR_256  = 0,
    MS8607_PRESSURE_RESOLUTION_OSR_512  = 1,
    MS8607_PRESSURE_RESOLUTION_OSR_1024 = 2,
    MS8607_PRESSURE_RESOLUTION_OSR_2048 = 3,
    MS8607_PRESSURE_RESOLUTION_OSR_4096 = 4,
    MS8607_PRESSURE_RESOLUTION_OSR_8192 = 5
} ms8607_pressure_resolution_t;

/* Operation parameters */
typedef enum ms8607_battery_status {
    MS8607_BATTERY_OK                   = 0,
    MS8607_BATTERY_LOW                  = 1
} ms8607_battery_status;

typedef enum ms8607_heater_status {
    MS8607_HEATER_OFF                   = 0,
    MS8607_HEATER_ON                    = 1
} ms8607_heater_status;

/******************************************************************************
Exported global functions (to be accessed by other files)
******************************************************************************/

void            R_MS8607_Init(void);
ms8607_status   R_MS8607_Reset(void);
ms8607_status   R_MS8607_Read_Prom(void);
ms8607_status   R_MS8607_Set_Humidity_Resolution(void);
ms8607_status   R_MS8607_Read_Temperature_Pressure_Humidity(float* temperature, float* pressure, float* relative_humidity);

#endif /* MS8607_H */
