/******************************************************************************
 * DISCLAIMER
 * This software is supplied by Renesas Electronics Corporation and is only
 * intended for use with Renesas products. No other uses are authorized. This
 * software is owned by Renesas Electronics Corporation and is protected under
 * all applicable laws, including copyright laws.
 * THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
 * THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
 * LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
 * TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS
 * ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE
 * FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
 * ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * Renesas reserves the right, without notice, to make changes to this
 * software and to discontinue the availability of this software. By using this
 * software, you agree to the additional terms and conditions found by
 * accessing the following link:
 * http://www.renesas.com/disclaimer
 *
 * Copyright (C) 2019 Renesas Electronics Corporation. All rights reserved.
 *****************************************************************************/
/*******************************************************************************
* File Name    : r_app_lora.c
* Version      : 1.0
* Description  : LoRaWAN controll application.
*                Connect with a LoRaWAN gateway to send sensor data.
******************************************************************************/
/*****************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 31.05.2019 1.00     First Release
******************************************************************************/

/******************************************************************************
Includes
******************************************************************************/

#include "ll.h"
#include "r_app_lora.h"
#include "r_app_ms8607.h"
#include "r_uart_lora.h"
#include "dataflash.h"
#include "vuarts.h"
#include "timer.h"
#include "r_cg_userdefine.h"

/******************************************************************************
Typedef definitions
******************************************************************************/

typedef struct lora_atcmd
{
    uint8_t no;
    uint8_t *cmdstr;
    uint8_t cmdlen;
} lora_atcmd_t;

/******************************************************************************
Macro definitions
******************************************************************************/

#define     RUN_COMMAND_QUEUE_SIZE      10

/******************************************************************************
Exported global variables (to be accessed by other files)
******************************************************************************/

LoRa_Param_Flag_t lora_param_flag = { 0 };
LoRa_Param_Flag_t lora_param_flagr = { 0 };

extern uint8_t app_mode;
uint8_t app_lora_param_f;

/* LoRa module AT commands */
const lora_atcmd_t lora_atcmd[] = {
    {0x00, 0x00, 0x00},/* dummy data */
    /* no, *cmdstr, cmdlen */
    {0x01, "AT+REGION=",  10},
    {0x02, "AT+DEVEUI=",  10},
    {0x03, "AT+CLASS=",   9},
    {0x04, "AT+APPEUI=",  10},
    {0x05, "AT+APPKEY=",  10},
    {0x06, "AT+ACTMODE=", 11},
    {0x07, "AT+FPORT=",   9},
};

/* AT command buffer */
uint8_t atcmd_buf[50];

/* running AT command */
uint8_t atcmd_run;

/* result of AT command */
uint8_t atcmd_bit;

/* LoRa module parameter flags */
LoRa_Param_Flag_t   lpf = {0};
/*
    uint8_t region  : 1;
    uint8_t deveui  : 1;
    uint8_t class   : 1;
    uint8_t appeui  : 1;

    uint8_t appkey  : 1;
    uint8_t actmode : 1;
    uint8_t fport   : 1;
    non
*/

/******************************************************************************
Private global variables and functions
******************************************************************************/

static uint16_t     RunCmd[RUN_COMMAND_QUEUE_SIZE] = {0};
static uint8_t      WrPtr;
static uint8_t      RdPtr;

static uint8_t      app_rxstr[50] = {0};
static uint8_t      wpos;

static uint8_t r_hex2char(uint8_t val);

extern void R_BLE_APP_Send_Execution_Result(void);

/******************************************************************************
* Function Name: R_LORA_APP_Init
* Description  : LoRa application initialize
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_LORA_APP_Init(void)
{
    if ((lpf.param & LORA_PARAM_FLAG_MASK) == LORA_PARAM_FLAG_MASK)
    {
        /* All LoRa module parameters are setted */
        app_lora_param_f = 1;
    }
    else
    {
        /* Insufficient LoRa module parameter setting     */
        /* Wait to be pushed the SW_USR       */
        /* for move to parameter setting mode */
        app_lora_param_f = 0;
    }

    /* Initialize Run Command Queue Pointer */
    WrPtr = 0;
    RdPtr = 0;

    R_LORA_APP_Set_RunCmd(LORA_AT_ECHO_DI_CMD);

    app_uart2_init();
}

/******************************************************************************
* Function Name: R_LORA_APP_Run
* Description  : Application Command Execute
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_LORA_APP_Run(void)
{
    /* Loop until Run Command Queue is empty */
    while(RdPtr != WrPtr)
    {
        switch(RunCmd[RdPtr])
        {
            case LORA_CMD_NON:
                /* do nothing */
                break;

            case LORA_AT_ECHO_DI_CMD:
                /* Disable echo-back */
                app_uart2_send("ATE0\r\n");
                break;

            case LORA_AT_PARAM_SEND_CMD:
                /* Send AT command to LoRaWANTM stack sample application board */
                app_uart2_send((char *)atcmd_buf);
                break;

            case LORA_AT_RESP_SEND_CMD:
                /* Send execution result of AT command to smartphone by BLE */
                RBLE_Set_Timer_Id(_RBLE_TIMER_ID_EXRES_IND,
                                  (RBLE_FUNC)R_BLE_APP_Send_Execution_Result,
                                  BLE_SEND_EXRESULT_INTERVAL);
                break;

            case LORA_AT_SAVE_CMD:
                /* Save parameters in LoRaWANTM stack sample application board */
                app_uart2_send("AT+SAVE\r\n");
                break;

            case LORA_AT_JOIN_CMD:
                /* Join LoRaWAN gateway */
                led_blink_interval = LED_BLINK_JOIN;
                RBLE_Set_Timer_Id(_RBLE_TIMER_ID_LED,
                                  (RBLE_FUNC)R_APP_LED2_Blink,
                                  led_blink_interval);
                app_uart2_send("AT+JOIN\r\n");
                atcmd_run = 0;
                break;

            case LORA_AT_SENDHEX_CMD:
                /* Send sensor data to LoRaWAN gateway */
                RBLE_Set_Timer_Id(_RBLE_TIMER_ID_SEND_SENSOR_DATA,
                                  (RBLE_FUNC)R_MS8607_APP_Start_Mesurement,
                                  LORAWAN_SEND_INTERVAL_1ST);
                break;

            default:
                break;
        }
        
        /* Operate Run Command Queue Pointer as a Ring Buffer */
        GLOBAL_INT_DISABLE();
        RdPtr = (RdPtr == (RUN_COMMAND_QUEUE_SIZE - 1)) ? (0) : (RdPtr + 1);
        GLOBAL_INT_RESTORE();
    }
}

/******************************************************************************
* Function Name: R_LORA_APP_Set_RunCmd
* Description  : cmd - Command value, defined by LORA_APP_CMD_enum
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_LORA_APP_Set_RunCmd(uint16_t cmd)
{
    RunCmd[WrPtr] = cmd;

    /* Operate Run Command Queue Pointer as a Ring Buffer */
    GLOBAL_INT_DISABLE();
    WrPtr = (WrPtr == (RUN_COMMAND_QUEUE_SIZE - 1)) ? (0) : (WrPtr + 1);
    GLOBAL_INT_RESTORE();
}

/******************************************************************************
* Function Name: R_LORA_APP_ATCmd_Generate
* Description  : This function converts the hexadecimal data received 
*                from the smartphone into AT commands.
* Arguments    : value - hexadecimal data
*                len   - hexadecimal data length
* Return Value : void
******************************************************************************/
void R_LORA_APP_ATCmd_Generate(int8_t *value, uint8_t len)
{
    int8_t  i;
    uint8_t wpos;

    /* copy AT command part(ex. "AT+REGION=") to buffer */
    memcpy(atcmd_buf, lora_atcmd[value[0]].cmdstr, lora_atcmd[value[0]].cmdlen);

    atcmd_run = value[0];   /* save AT command number */
    atcmd_bit = (uint8_t)(1 << (atcmd_run - 1));
                            /* save AT command bit */

    wpos = lora_atcmd[value[0]].cmdlen;
    for(i = 1; i < len; i++)
    {
        /* Exchange from hexadecimal to ascii character */
        
        if ((atcmd_run == 0x03) || (atcmd_run == 0x06))
        {
            /* "AT+CLASS=" or "AT+ACTMODE=" */
            atcmd_buf[wpos++] = r_hex2char(value[i] & 0x0F);
        }
        else
        {
            atcmd_buf[wpos++] = r_hex2char((value[i]>>4) & 0x0F);
            atcmd_buf[wpos++] = r_hex2char(value[i] & 0x0F);
        }
    }
    atcmd_buf[wpos++] = '\r';
    atcmd_buf[wpos++] = '\n';
    atcmd_buf[wpos]   = 0x00;
}

/******************************************************************************
* Function Name: R_LORA_APP_ATResult_Generate
* Description  : Judgment and processing of execution results in LoRaWAN module.
* Arguments    : c - Ascii character. It is received from LoRaWAN module.
* Return Value : void
******************************************************************************/
void R_LORA_APP_ATResult_Generate(uint8_t c)
{
    app_rxstr[wpos] = c;

    /* Detects CR and LF */
    if ((app_rxstr[wpos] == '\n') && (app_rxstr[wpos - 1] == '\r'))
    {
        /* When echo-back is disabled, the string "ATE0" is received first, but it is ignored. */
        
        if (strstr((const char *)app_rxstr, (const char *)"OK") != NULL)
        {
            if (app_mode == APP_MODE_INITIALIZATION)
            {
                /* OK of ATE0 command */

                if (app_lora_param_f == 1)
                {
                    R_LORA_APP_Set_RunCmd(LORA_AT_JOIN_CMD);
                    R_APP_Mode_Set(APP_MODE_LORA_JOIN);
                }
                else
                {
                    /* Wait for be pushed button after ATE0 command */
                    R_LORA_APP_Set_RunCmd(LORA_CMD_NON);
                    R_APP_Mode_Set(APP_MODE_IDLE);
                }
            }
            else
            if (app_mode == APP_MODE_LORA_SAVE_PARAM)
            {
                R_LORA_APP_Set_RunCmd(LORA_AT_JOIN_CMD);
                R_APP_Mode_Set(APP_MODE_LORA_JOIN);
            }
            else
            if (app_mode == APP_MODE_LORA_JOIN)
            {
                /* OK of AT+JOIN command */

                /* wait for "+JOIN" response */
            }
            else
            if (app_mode == APP_MODE_BLE_PARAM_SETTING)
            {
                /* OK of LoRa parameter setting commands */
                lpf.param |= atcmd_bit;
                atcmd_buf[1] = (uint8_t)true;
            }
            else
            if (strstr((const char *)app_rxstr, (const char *)"+SENDHEX") != NULL)
            {
                /* do nothing */
            }
            else
            {
                /* Wait to receive JOIN_ACCEPTED or JOIN_FAILED */
                /* do nothing here */
            }
        }
        else
        if (strstr((const char *)app_rxstr, (const char *)"ERROR") != NULL)
        {
            if (app_mode == APP_MODE_BLE_PARAM_SETTING)
            {
                /* LoRa parameter setting command error */
                lpf.param &= ~atcmd_bit;
                atcmd_buf[1] = (uint8_t)false;
            }
            else
            {
                /* JOIN error */
            }

            /* LoRa command error */
            LED2 = LED_ON;
        }
        else
        if (strstr((const char *)app_rxstr, (const char *)"+JOIN") != NULL)
        {
            if (strstr((const char *)app_rxstr, (const char *)"JOIN_ACCEPTED") != NULL)
            {
                led_blink_interval = LED_BLINK_NORMAL;
                R_LORA_APP_Set_RunCmd(LORA_AT_SENDHEX_CMD);
                R_APP_Mode_Set(APP_MODE_LORA_COMMUNICATION);
            }
            else
            {
                RBLE_Clear_Timer_Id(_RBLE_TIMER_ID_LED);
                
                /* JOIN_FAILED */
                LED2 = LED_ON;
            }

            /* Release push processing of SW_USR */
            sw_usr_is_pushed = SW_RELEASED;
        }
        else
        if (strstr((const char *)app_rxstr, (const char *)"+RCVD") != NULL)
        {
            int8_t *str;
            
            /* Receive a command at downlink */

            str = (int8_t *)strtok((char *)app_rxstr, (const char *)" ");   /* find +RCVD: */
            str = (int8_t *)strtok((char *)NULL, (const char *)",");        /* find <data> */
            
            if (strcmp((const char *)str, (const char *)LORAWAN_DOWNLINK_STRING) == (int)NULL)
            {
                /* When a LoRa parameter setting command is received at the downlink, */
                /* exit from LoRa transmission mode.                                  */

                RBLE_Clear_Timer_Id(_RBLE_TIMER_ID_SEND_SENSOR_DATA);
                RBLE_Clear_Timer_Id(_RBLE_TIMER_ID_LED);
                LED2 = LED_OFF;

                R_APP_Mode_Set(APP_MODE_IDLE);
                R_LORA_APP_Set_RunCmd(LORA_CMD_NON);
            }
        }
        else
        {
            /* Only CR+LF was received */
        }

        /* CR+LF was found. and clear buffer. */
        wpos = 0;
        memset(app_rxstr, 0, sizeof(app_rxstr));

        if (app_mode == APP_MODE_BLE_PARAM_SETTING)
        {
            /* Send AT command result by ble to smartphone */

            atcmd_buf[0] = atcmd_run;   /* current AT command */
            atcmd_buf[2] = lpf.param;   /* it has been set commands */
            
            R_LORA_APP_Set_RunCmd(LORA_AT_RESP_SEND_CMD);
        }
    }
    else
    {
        wpos++;
    }
}

/******************************************************************************
* Function Name: r_hex2char
* Description  : Exchange from hexadecimal to ascii code character
* Arguments    : val - Hexadecimal
* Return Value : c - Ascii code character
******************************************************************************/
static uint8_t r_hex2char(uint8_t val)
{
    uint8_t c;

    if (val <= 9)
    {
        /* 0 - 9 */
        c = val + 0x30;
    }
    else
    if ((val >= 0xA) && (val <= 0xF))
    {
        /* A - F */
        c = val + 0x37;
    }
    else
    {
        /* do nothing */
    }

    return c;
}
