/******************************************************************************
 * DISCLAIMER
 * This software is supplied by Renesas Electronics Corporation and is only
 * intended for use with Renesas products. No other uses are authorized. This
 * software is owned by Renesas Electronics Corporation and is protected under
 * all applicable laws, including copyright laws.
 * THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
 * THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT
 * LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED.
 * TO THE MAXIMUM EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS
 * ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES SHALL BE LIABLE
 * FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR
 * ANY REASON RELATED TO THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE
 * BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * Renesas reserves the right, without notice, to make changes to this
 * software and to discontinue the availability of this software. By using this
 * software, you agree to the additional terms and conditions found by
 * accessing the following link:
 * http://www.renesas.com/disclaimer
 *
 * Copyright (C) 2019 Renesas Electronics Corporation. All rights reserved.
 *****************************************************************************/
/*******************************************************************************
* File Name    : r_app_ms8607.c
* Version      : 1.0
* Description  : MS8607 sensor controll application.
******************************************************************************/
/*****************************************************************************
* History : DD.MM.YYYY Version  Description
*         : 31.05.2019 1.00     First Release
******************************************************************************/

/******************************************************************************
Includes
******************************************************************************/

#include    "r_cg_macrodriver.h"
#include    "timer.h"
#include    "ll.h"
#include    "rble.h"
#include    "vuarts.h"
#include    "r_uart_lora.h"
#include    "r_app_ms8607.h"
#include    "r_ms8607.h"
#include    "r_cg_userdefine.h"

/******************************************************************************
Macro definitions
******************************************************************************/

#define     RUN_COMMAND_QUEUE_SIZE          10

#define     MS8607_MESUREMENT_INTERVAL      500 /* 500 * 10ms = 5s */

/******************************************************************************
Typedef definitions
******************************************************************************/

/* MS8607 Application Run Sequence Commands */
enum MS8607_APP_CMD_NO_enum {
    MS8607_APP_CMD_NON,
    MS8607_APP_RESET_CMD,
    MS8607_APP_START_MEASUREMENT_CMD,
};

/* MS8607 Application Processing State */
enum MS8607_APP_STATE_NO_enum {
    MS8607_APP_IDLE_STATE,
    MS8607_APP_INITIALIZE_STATE,
    MS8607_APP_MEASURING_STATE,
    MS8607_APP_STOP_STATE,
};

/* General Purpose Communication Server Data Structure */
typedef struct GPCS_data {
    uint8_t rbuf[VUART_SEND_BUFFER_SIZE];   /* Receive data buffer of BLE write request. */
    uint8_t sbuf[VUART_SEND_BUFFER_SIZE];   /* Send data buffer of BLE indication. */
    uint8_t rlen;                           /* Receive data length. */
    uint8_t slen;                           /* Send data length. */
    uint8_t write_req;                      /* Write request flag. Set by the RBLE_VUART_EVENT_SERVER_WRITE_REQ event. */
    uint8_t indication_cfm;                 /* Indication confirmation flag. Not used. */
    uint8_t state;                          /* Processing State. */
} st_GPCS_data_t;

/******************************************************************************
Exported global variables (to be accessed by other files)
******************************************************************************/

/* Sensor mesurement variables */
float Temperature;
float Pressure;
float Relative_Humidity;

/* General Purpose Communication Server Data Structure */
st_GPCS_data_t  GPCS_data;

int8_t          sensor_data_str[50];

/******************************************************************************
Private global variables and functions
******************************************************************************/

static void r_MS8607_APP_IIC_Transfer_Error(void);

/******************************************************************************
* Function Name: R_MS8607_APP_Init
* Description  : Initialize the sensor application.
* Arguments    : void 
* Return Value : void
******************************************************************************/
void R_MS8607_APP_Init(void)
{
    R_MS8607_Init();
    R_MS8607_APP_Reset();
    return;
}

/******************************************************************************
* Function Name: R_MS8607_APP_Reset
* Description  : Reset function.
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_MS8607_APP_Reset(void)
{
    ms8607_status   err;

    err = R_MS8607_Reset();
    if(err != MS8607_STATUS_OK)
    {
        r_MS8607_APP_IIC_Transfer_Error();
    }

    err = R_MS8607_Read_Prom();
    if(err != MS8607_STATUS_OK)
    {
        r_MS8607_APP_IIC_Transfer_Error();
    }

    err = R_MS8607_Set_Humidity_Resolution();
    if(err != MS8607_STATUS_OK)
    {
        r_MS8607_APP_IIC_Transfer_Error();
    }
}

/******************************************************************************
* Function Name: R_MS8607_APP_Start_Mesurement
* Description  : Measure temperature, pressure and relative humidity.
* Arguments    : void
* Return Value : void
******************************************************************************/
void R_MS8607_APP_Start_Mesurement(void)
{
    ms8607_status   err;
    uint32_t        tm, pr, hu;
    static uint8_t  data_f = 0;

    /* Measure temperature, pressure and relative humidity. */
    err = R_MS8607_Read_Temperature_Pressure_Humidity(&Temperature,
                                                      &Pressure,
                                                      &Relative_Humidity);
    if(err != MS8607_STATUS_OK)
    {
        r_MS8607_APP_IIC_Transfer_Error();
    }

    memcpy(&tm, &Temperature, sizeof(uint32_t));
    memcpy(&pr, &Pressure, sizeof(uint32_t));
    memcpy(&hu, &Relative_Humidity, sizeof(uint32_t));

    if (data_f == 0)
    {
        /* Temperature */
        sprintf((char *)sensor_data_str, "AT+SENDHEX=%04x%04x\r\n",
                (uint16_t)(tm>>16),
                (uint16_t)tm
               );
    }
    else
    {
        /* Pressure & Relative_Humidity */
        sprintf((char *)sensor_data_str, "AT+SENDHEX=%04x%04x%04x%04x\r\n",
                (uint16_t)(pr>>16),
                (uint16_t)pr,
                (uint16_t)(hu>>16),
                (uint16_t)hu
               );
    }
    data_f ^= 1;
    app_uart2_send((char *)sensor_data_str);

    /* Call this function again. (Stops when SW_USR is pushed.) */
    RBLE_Set_Timer_Id(_RBLE_TIMER_ID_SEND_SENSOR_DATA,
                      (RBLE_FUNC)R_MS8607_APP_Start_Mesurement,
                      LORAWAN_SEND_INTERVAL);
    return;
}

/******************************************************************************
* Function Name: r_MS8607_APP_IIC_Transfer_Error
* Description  : Transfer error function of IIC.
* Arguments    : void
* Return Value : void
******************************************************************************/
static void r_MS8607_APP_IIC_Transfer_Error(void)
{
    uint32_t    i;

    DI();

    LED1 = LED_ON;
    LED2 = LED_OFF;
    while(1)
    {
        LED1 ^= 1;
        LED2 ^= 1;
        i = 200000;
        while(i--);
    }
}
